---
title: Can I haz ur dataz
categories:
  - cyber
  - brain fart
tags:
date: 2019-02-01 09:55:07
---

Remote data acquisitions over ssh
---------------------------------

To get your data trough `ssh` to your local storage, you simply use pipes. It does not matter if you use `cat`, `dd` or any other command line tool which outputs the data on standard output (_stdout_).

### Using `cat`

When using cat the there is no need to add any additional parameters in your command chain. A simple `cat <input>` will suffice.  
  
`Cat` does not have any progress information during read operations.

### Using `dd`

The program `dd` requires more user interaction during the setup phase. To use `dd` you have to give the _if=_ argument. The use of different blocksizes (_bs_) will not have that much of an impact on the speed. Feel free to have a look at this [this](https://goatpr0n.farm/2019/01/23/what-if-the-cult-of-dd-is-right/).

![](https://dl.goatpr0n.events/$/VAN9L)

_﻿Wow. So scientific! Much CLI._

Examples
--------

A simple example with no output to the terminal except of errors. The output to _stderr_ is not suppressed.

``` bash
$ # Using cat to copy /dev/sda
$ ssh <user@remotehost> 'cat /dev/sda' > sda.raw
```

If you want to suppress errors, use:

``` bash
$ # Using cat to copy /dev/sda
$ ssh <user@remotehost> 'cat /dev/sda 2>/dev/null' > sda.raw
```

The next example will demonstrate the use of `dd`.

``` bash
$ # Using dd to copy /dev/sda
$ ssh <user@remotehost> 'dd if=/dev/sda' > sda.raw
```

Of course you can suppress errors as well.

Speeding up the data acquisition and minor tweaks
-------------------------------------------------

With the basics covered, we can begin optimizing our data transfer. In the first step we speed up the transfer with `gzip`.  
  
The argument _-c_ will write the compressed data to _stdout_ which will be piped to your local system.  
  
Of course you can use this with `cat` as well.

``` bash
$ ssh <user@remotehost> 'dd if=/dev/sda | gzip -c' | gunzip > sda.raw
```

To add some progress information, you have two options with `dd`.

``` bash
$ # dd status
$ ssh <user@remotehost> 'dd if=/dev/sda status=progress | gzip -c' \
> | gunzip > sda.raw
```

The _status_ argument writes the output to _stderr_ and will not end up in your local copy.

``` bash
$ # dd through pv
$ ssh <user@remotehost> 'dd if=/dev/sda | gzip -c' \
> | pv | gunzip > sda.raw
```

`Pv` needs to be installed separatly. Check your packet manager. 0r teh Googlez!

KTHXBYE.

#### Update #01

Fixed a problem within the examples. Had a pipe too much. #Fail
